#include "proj.h"
#define DEBUG
#ifndef DEBUG
  #define eprintf( fmt, args... )
  #define DBGOUT(...)
#else
  #if DEBUGHW>0
    #define FOO(...) __VA_ARGS__
    #define DBGOUT dbg_string+= FOO
    #if (DEBUGHW==2)
      #define eprintf(fmt, args...) S.printf(fmt, ##args)
    #elif (DEBUGHW==1 || DEBUGHW==3)
      #define eprintf(fmt, args...) sprintf(dbg,fmt, ##args);dbg_string+=dbg;dbg[0]=0
    #endif
  #else
    #define eprintf( fmt, args... )
    #define DBGOUT(...)
  #endif
#endif

void secTick();
#if DEBUGHW==1
  WiFiServer dbg_server(10000);
  WiFiClient dbg_client;
#endif
#ifdef STROMPREIS
String strompreis="";
#endif // strompreis
Ticker uniTicker,secTicker;
strConfig config;
AsyncMqttClient mq_client;                    // ThingsPeak Client
unsigned things_cycle;
uint32_t a_result[9];
uint8_t key[16];
String things_up;
unsigned thingspeak_watch;
bool new_data,new_data3,ledbit,ledflag;
unsigned first_frame;
uint8_t dow_local,dow;
unsigned kwh_day_in[7];
unsigned kwh_day_out[7];
uint32_t clientId;
unsigned prev_millis;
int logPage=-1;
#if DEBUGHW>0
  char dbg[128];
  String dbg_string;
#endif // DEBUGHW
kwhstruct kwh_hist[7];
bool inAPMode,mqttStatus,hwTest;
ADC_MODE(ADC_VCC);

void thingspeak_publish() {
  String data = "";
#ifdef STROMPREIS
  for (unsigned i=0;i<7;i++)
    data += "&field" + (String(i+1))+"="+(String)(a_result[i]);
    data += "&field8="+strompreis;
#else
  for (unsigned i=0;i<8;i++)
    data += "&field" + (String(i+1))+"="+(String)(a_result[i]);
#endif // strompreis
  String topicString ="channels/" + (String)config.channel_id + "/publish/"+String(config.write_api_key);
  mq_client.publish( topicString.c_str(),0,false, data.c_str() );
  things_up=timecode;
  thingspeak_watch=0;
//  DBGOUT("ThingsPeak publish\n");
//  DBGOUT(topicString+"\n");
//  DBGOUT(data+"\n");
}

void onMqConnect(bool sessionPresent) {
  DBGOUT("Thingspeak connected\n");
  thingspeak_publish();
}

void onMqDisconnect(AsyncMqttClientDisconnectReason reason) {
  DBGOUT("ThingsPeak Disconnected from MQTT.\n");
}

void setup(){
//  char stack;
//  stack_start = &stack;
  Serial.begin(9600,SERIAL_8E1);      // Schnittstelle zu Amis-Z?er
  Serial.setTimeout(10);              // f. readBytes in amis.cpp
  pinMode(AP_PIN,INPUT_PULLUP);
  #if LEDPIN
  digitalWrite(LEDPIN,HIGH);
  pinMode(LEDPIN, OUTPUT);
  #endif // if LEDPIN == Serial1.txd: reroute pin function
  #if DEBUGHW==2
    #if DEBUG_OUTPUT==0
      Serial.begin(115200);
    #elif DEBUG_OUTPUT==1
      Serial1.begin(115200);
    #endif
  #endif // DEBUGHW
  //Serial1.setDebugOutput(true);
  //S.print("\nStart\n");
  #ifdef OTA
  initOTA();
  #endif // OTA
  LittleFS.begin();                 // always true! SPIFF.begin does autoformat!!!
  bool test=false;
  File f = LittleFS.open("/index.html", "r");
  if(f) f.close();
  else test=true;
  if (!test) {
    f = LittleFS.open("/custom.css", "r");
    if(f) f.close();
    else test=true;
  }
  if (test) {                     // keine html-Files
    serverInit(1);                // /upgrade.html als /
    upgrade(0);
    return;
  }
  serverInit(0);                  // /init.html als /
  generalInit();
  histInit();
  if (config.thingspeak_aktiv) {
    mq_client.onConnect(onMqConnect);
    mq_client.onDisconnect(onMqDisconnect);
    mq_client.setServer("mqtt.thingspeak.com", 1883);
  }
  connectToWifi();  // and MQTT and NTP
  secTicker.attach_scheduled(1,secTick);
  if (config.log_sys) writeEvent("INFO", "sys", "System setup completed, running", "");
}

void loop(){
  #if DEBUGHW==1
  if (dbg_string.length()) {          // Debug-Ausgaben TCP
    dbg_string+="\n";
    if (!dbg_client.connected()) dbg_client.stop();
    if (!dbg_client) dbg_client = dbg_server.available();
    if (dbg_client)  dbg_client.print(dbg_string);
    dbg_string="";
  }
  #elif DEBUGHW==2
  if (dbg_string.length()) {          // Debug-Ausgaben Serial
    S.print(dbg_string);
    dbg_string="";
  }
  #elif DEBUGHW==3
  if (dbg_string.length()) {          // Debug-Ausgaben Websock
    ws.text(clientId,dbg_string);
    //Serial1.println(dbg_string);
    dbg_string="";
  }
  #endif

  #ifdef OTA
  ArduinoOTA.handle();
  #endif
  if(shouldReboot){
    secTicker.detach();
    mqttTimer.detach();
    if (config.log_sys) writeEvent("INFO", "sys", "System is going to reboot", "");
    DBGOUT("Rebooting...");
    delay(300);
    ESP.wdtDisable();           // bootet 2x ???
    //ESP.restart();
    while (1)    delay(1);
  }
  if (config.thingspeak_aktiv && thingspeak_watch>10) {
    if (config.log_sys) writeEvent("INFO", "mqtt", "Connection lost long time", "reboot");
    DBGOUT("thingspeak_watch reset");
    shouldReboot=true;
  }
  if (ws.count()) {      // ws-connections
    if (new_data3) {
      new_data3=false;
      sendZData();
    }
  }
  if (WiFi.isConnected())   amis_poll();
  #ifdef LEDPIN
  if (inAPMode) {
    if (millis() > prev_millis) {
      prev_millis=millis()+500;
      digitalWrite(LEDPIN,digitalRead(LEDPIN)^1);
    }
  }
  else {
    if (ledflag && ledbit) {
        digitalWrite(LEDPIN,LOW);
    }
  }
  #endif
  if (hwTest) {
    for (unsigned i=0;i < 200; i++)  {
      Serial.write(i);
      delay(1);
    }
  }
  if (logPage >=0) {
    sendEventLog(clientId,logPage);
    logPage=-1;
  }
  delay(10);
  #ifdef LEDPIN
  if (ledflag) {
    ledflag=false;
    digitalWrite(LEDPIN,HIGH);
  }
  #endif
}

void writeHistFileIn(int x, long val) {
  DBGOUT("hist_in "+String(x)+" "+String(val)+"\n");
  File f = LittleFS.open("/hist_in"+String(x), "w");
  if(f) {
    f.print(val);
    f.close();
  }
}
void writeHistFileOut(int x, long val) {
  DBGOUT("hist_out "+String(x)+" "+String(val)+"\n");
  File f = LittleFS.open("/hist_out"+String(x), "w");
  if(f) {
    f.print(val);
    f.close();
  }
}

void secTick() {
  things_cycle++;
  #ifdef LEDPIN
  if (things_cycle % 4==0) ledflag=true;
  #endif
  if (ws.count()) {        // ws-connections
    if (first_frame==0) {
      sendZDataWait();
    }
  }
  if (first_frame==3) {
    first_frame=2;
    int x=dow-2;        // gestern
    if (x < 0) x=6;
    if (x>6) x=0;
    if (kwh_day_in[x] ==0) {    // gestern noch keine Werte: momentanen Stand wegschreiben
      kwh_day_in[x]=a_result[0];
      writeHistFileIn(x,a_result[0]);
    }
    if (kwh_day_out[x] ==0) {    // gestern noch keine Werte: momentanen Stand wegschreiben
      kwh_day_out[x]=a_result[1];
      writeHistFileOut(x,a_result[1]);
    }
    dow_local=dow;
  }
  else if (first_frame==2) {
    int x=dow-2;                    // gestern, idx ab 0
    if (x < 0) x=6;
    for (int i=0; i<7;i++) {
      int vg=x-1; if (vg <0) vg=6;
      if ((kwh_day_in[x] > kwh_day_in[vg]) && kwh_day_in[vg]) {
        kwh_hist[i].kwh_in = kwh_day_in[x] - kwh_day_in[vg];
        kwh_hist[i].dow = x;
      }
      else kwh_hist[i].kwh_in = 0;
      if ((kwh_day_out[x] > kwh_day_out[vg]) && kwh_day_out[vg]) {
        kwh_hist[i].kwh_out = kwh_day_out[x] - kwh_day_out[vg];
        kwh_hist[i].dow = x;
      }
      else kwh_hist[i].kwh_out = 0;
      x--; if (x <0) x=6;
    }
    histUpdate();
    first_frame=1;                  // state inaktiv
  }
  if (dow_local != dow) {           // Tageswechsel, dow 1..7
    int x=dow-2;                    // gestern, idx ab 0
    if (x < 0) x=6;                 // x zeigt auf gestern
    kwh_day_in[x]=a_result[0];
    writeHistFileIn(x,a_result[0]);
    kwh_day_out[x]=a_result[1];
    writeHistFileOut(x,a_result[1]);
    first_frame=2;
    dow_local=dow;
    histUpdate();
  }
  if (config.thingspeak_aktiv && things_cycle >= config.thingspeak_iv && new_data) {
    things_cycle=0;
    thingspeak_watch++;
    new_data = false;
    // Thingspeak trennt nach ca 10 sec, Verbindung muss jedesmal neu aufgebaut werden
    // Au?rdem: Thingspeak akzeptiert JEDEN WriteApiKey, nur kommen die Daten natürlich nicht an
    // Die PubSubClient.h muss mod. werden, damit der lange Payload-String durchgeht: #define MQTT_MAX_PACKET_SIZE 256
    mq_client.setCredentials("",config.write_api_key.c_str());
    //mq_client.setMaxTopicLength(256);    // ~ 100
    if (!mq_client.connected()) mq_client.connect();  // and publish
    else thingspeak_publish();
  }
  ws.cleanupClients();
}

void  writeEvent(String type, String src, String desc, String data) {
  DynamicJsonBuffer jsonBuffer;
  JsonObject &root = jsonBuffer.createObject();
  root[F("type")] = type;
  root[F("src")] = src;
  root[F("desc")] = desc;
  root[F("data")] = data;
  root[F("time")] = timecode;
  File eventlog = LittleFS.open("/eventlog.json", "a");
  if(eventlog.size() > 50000) {
    eventlog.close();
    LittleFS.remove("/eventlog.json");
    File eventlog = LittleFS.open("/eventlog.json", "a");
//    writeEvent("WARN", "sys", "Event log cleared!", "");
  }
  root.printTo(eventlog);
  eventlog.print("\n");
  eventlog.close();
}
